#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <pthread.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "hashtable.h"
#ifdef WITH_DMALLOC
#include <dmalloc.h>
#endif

typedef struct hashentry_struct {
  struct hashentry_struct *next;
  unsigned int hashkey;
  const void *key;
  const void *value;
} hashentry;

struct hashtable_struct {
  hashentry **table;
  unsigned int table_size;
  unsigned int (*hash)(const void *key);
  boolean (*isequal)(const void *key1, const void *key2);
  /*@null@*/ void (*free_key)(void *key);
  /*@null@*/ void (*free_value)(void *value);
#ifdef _REENTRANT
  pthread_mutex_t mutex;
#endif
};

hashtable ht_new(unsigned int size, 
                 unsigned int (*hash)(const void *key),
                 boolean (*isequal)(const void *key1, 
                                    const void *key2),
                 void (*free_key)(void *key),
                 void (*free_value)(void *value)) {
  hashtable ht = malloc(sizeof(struct hashtable_struct));
  assert(ht!=NULL);
  ht->table = (hashentry **)malloc(sizeof(hashentry *)*size);
  assert(ht->table!=NULL);
  memset(ht->table, 0, size * sizeof(hashentry *));   
  ht->table_size = size;
  ht->hash = hash;
  ht->isequal = isequal;
  ht->free_key = free_key;
  ht->free_value = free_value;
#ifdef _REENTRANT
  (void)pthread_mutex_init(&ht->mutex, NULL);
#endif
  return(ht);
}

#ifdef _REENTRANT
#define LOCK assert(pthread_mutex_lock(&ht->mutex)==0);
#define UNLOCK assert(pthread_mutex_unlock(&ht->mutex)==0);
#else
#define LOCK 
#define UNLOCK
#endif

void ht_insert(hashtable ht, 
               const void *key,
               const void *value) {
  hashentry *he;
  assert(ht!=NULL);
  he = malloc(sizeof(hashentry));
  assert(he!=NULL);
  he->hashkey = ht->hash(key);
  he->key = key;
  he->value = value;
  LOCK;
  he->next = ht->table[he->hashkey%ht->table_size];
  ht->table[he->hashkey%ht->table_size] = he;
  UNLOCK;
}

void ht_free_entry(hashtable ht, 
                   const void *key,
                   const void *value) {
  if(ht->free_key!= NULL)
    ht->free_key((void *)key);
  if(ht->free_value!= NULL)
    ht->free_value((void *)value);
}
void ht_delete(/*@only@*/ hashtable ht) {
  unsigned int i;
  LOCK;
  for(i=0; i<ht->table_size; i++) {
    hashentry *he = ht->table[i];
    hashentry *prvhe;
    while(he) {
      ht_free_entry(ht, he->key, he->value);
      prvhe = he;
      he=he->next;
      free(prvhe);
    }
  }
  UNLOCK;
  free(ht->table);
  free(ht);
}

void ht_remove(hashtable ht, const void *key) {
  unsigned int hkey;
  hashentry *he, *prehe;
  assert(ht!=NULL);
  hkey = ht->hash(key);
  prehe = NULL;
  LOCK;
  he = ht->table[hkey%ht->table_size];
  while(he!=NULL) {
    if(he->hashkey == hkey && ht->isequal(he->key,key)) {
      if(prehe != NULL) {
        prehe->next = he->next;
      } else {
        ht->table[hkey%ht->table_size] = he->next;
      }
      UNLOCK;
      free(he);
      return;
    }
    prehe = he;
    he = he->next;
  }
  UNLOCK;
  return;
}
const void *ht_lookup(hashtable ht,
                      const void *key) {
  unsigned int hkey;
  hashentry *he;
  assert(ht!=NULL);
  hkey = ht->hash(key);
  LOCK;
  he = ht->table[hkey%ht->table_size];
  while(he!=NULL) {
    if(he->hashkey == hkey && ht->isequal(he->key,key)) {
      UNLOCK;
      return(he->value);
    }
    he = he->next;
  }
  UNLOCK;
  return NULL;
}

boolean ht_iterate(hashtable ht,
                boolean (*callback)(void *user, 
                                 const void *key, 
                                 const void *val), 
                void *user) {
  unsigned int i;
  boolean cont = TRUE;
  LOCK;
  for(i=0; i<ht->table_size && cont; i++) {
    hashentry *he = ht->table[i];
    while(he) {
      cont = callback(user, he->key, he->value);
      he=he->next;
    }
  }
  UNLOCK;
  return cont;
}

